(**** Certified Dynamic Storage Allocation -Definition ***)

Require Import Arith.
Require Import heap2.
Require Import scap.
Require Import scap_rules.
Require Import List.
(*** shorthands for Assertions ****)
Definition getVinH (s : state) (i : addr) : nat :=
  match (getValinH s i) with
    | Some x => x
    | None => O
  end.


Notation pnil := O.
(** assertion land ***)
Definition Aand (a1 a2 : assert) : assert :=
  fun s : state => a1 s /\ a2 s.

(** assertion lor ****)
Definition Aor (a1 a2 : assert) : assert :=
  fun s : state => a1 s \/ a2 s.

(** empty heap *)
Definition emp : assert := 
  fun s : state => forall i : addr, getValinH s i = hvnone.

(** singleton heap *)
Definition pto (i : addr) (v : nat) : assert :=
  fun s : state =>
    getValinH s i = hvsome v /\
    forall j : addr, ~i = j -> getValinH s j = hvnone.

(* separation conjunction *)
Definition star (a1 a2 : assert) : assert :=
  fun s : state =>
    let (h, rf) := s in
      exists h1 : heap, exists h2 : heap,
	hdisj h1 h2 /\ heap_eq (happend h1 h2) h
	/\ a1 (h1, rf) /\ a2 (h2, rf).

Definition star1 (i : addr) (v : nat) (a : assert) : assert :=
  fun s : state =>
    getValinH s i = (hvsome v) /\ a (removeS s i).

(** separating implication **)
(* arbitrary heap *)
Definition arbh : assert :=
  fun s : state => True.

Definition simp (a1 a2 : assert) : assert :=
  fun s : state =>
    let (h, rf) := s in
      forall h' : heap, hdisj h h' -> a1 (h', rf) -> a2 (happend h h', rf).

(** singleton heap with unkown content **)
Definition ptoany (i : addr) : assert :=
  fun s : state => exists x : nat, (pto i x) s.

(** multi elem. heap **)
Fixpoint ptolist (a : addr) (l : list int) {struct l} : assert :=
  match l with
    | nil => emp
    | cons i l' => (star1 a i (ptolist (S a) l'))
  end.

Definition ptoanylist (a : addr) : assert :=
  fun s : state => exists l : list int, (ptolist a l) s.

Fixpoint ptoanynlist (a : addr) (n : int) {struct n} : assert :=
  match n with
    | O => emp
    | S n' => (star (ptoany a) (ptoanynlist (S a) n'))
  end.

Definition hindomS (i : addr) (s : state) : Prop :=
  let (h, rf) := s in
    hindom i h.

Inductive hsegindomS : addr -> int -> state -> Prop :=
  | all_in_heap :
    forall (base : addr) (len : int) (s : state),
      hindomS base s ->
      (hsegindomS (S base) (len - 1) s) ->
      hsegindomS base len s.

Definition Stk (bottom : addr) : assert :=
  ptoanynlist 1 bottom. 

(** _______p_____________________p+n-3_  **)
(**| q | s |                       |   | **)
(** -----------------------------------  **)
Definition MBlk (p q : addr) (n : nat) : assert :=
  fun s : state => p > 2 /\ n > 2 /\ (
    star1 (p - 2) q (
    star1 (p - 1) n (ptoanynlist p (n - 2))) s).

(** p________   p'________          __________  **)
(** |p'|*|   |   | | |    | .......|q|        | **)
(**  --------     --------          ----------  **)
Inductive MBlkLst : nat -> addr -> addr -> assert :=
  | mbl_O : forall (s : state) (p : addr),
    emp s -> MBlkLst O p p s
  | mbl_S : forall (h h1 h2 : heap) (rf : rfile) (n m : nat) (p p' q : addr),
    (p < p' \/ p' = pnil) ->
    hdisj h1 h2 ->
    heap_eq (happend h1 h2) h ->
    MBlk (p + 2) p' n (h1, rf) ->
    MBlkLst m p' q (h2, rf) ->
    MBlkLst (S m) p q (h, rf).

(** fl_________         __________     p_________  **)
(**  |  |  |   | ..... |p |       |    |q |      | **)
(**   ---------         ----------      ---------  **)
Definition EndL (fl p q : addr) : assert :=
  fun s : state =>
    (p = pnil -> MBlkLst O fl q s) /\
    (p <> pnil -> 
      ((p < q \/ q = pnil) /\ 
	(exists n, exists m, star (MBlkLst m fl p) (MBlk (p + 2) q n) s))).

(** fl_________         __________     p_________       __________   **)
(**  |  |  |   | ..... |p |       |    |q |      |...  |nil|       | **)
(**   ---------         ----------      ---------       -----------  **)
Definition MidL (fl p q : addr) : assert :=
  fun s : state =>
    exists n : nat, (star (EndL fl p q) (MBlkLst n q pnil)) s.

(* good free list is a MBlkLst *)
Definition Good (fl : addr) : assert :=
  fun s : state =>
    exists n : nat, (MBlkLst n fl pnil) s.

(**short hands for guarantee ***)
Definition RIntact (r : reg) : action :=
  fun s : state =>
    fun s' : state =>
      getValinR s r = getValinR s' r.

Definition Gand (g1 g2 : action) : action :=
  fun s : state =>
    fun s' : state =>
      g1 s s' /\ g2 s s'.

Definition HIntact (top bottom : addr) : action :=
  fun s : state =>
    fun s' : state => forall i : addr, i <= top /\ i > bottom ->
      getValinH s i = getValinH s' i.

Definition RfInsA (a : assert) : Prop :=
  forall (h : heap) (rf rf' : rfile), a (h, rf) -> a (h, rf').

(* intact callee saved registers **)
Definition S0S1S2 : action := Gand (
  RIntact s0)                (Gand (
  RIntact s1)                (Gand (
  RIntact s2)                (Gand (
  RIntact s3)                (Gand (
  RIntact s4)                (Gand (
  RIntact s5)                (Gand (
  RIntact s6)                (
  RIntact s7))))))).

Definition ES0 : action := Gand (
  RIntact s1)                (Gand (
  RIntact s2)                (Gand (
  RIntact s3)                (Gand (
  RIntact s4)                (Gand (
  RIntact s5)                (Gand (
  RIntact s6)                (
  RIntact s7)))))).

Definition ES0S1 : action := Gand (
  RIntact s2)                (Gand (
  RIntact s3)                (Gand (
  RIntact s4)                (Gand (
  RIntact s5)                (Gand (
  RIntact s6)                (
  RIntact s7))))).

Definition ES0S1S2 : action := Gand (
  RIntact s3)                (Gand (
  RIntact s4)                (Gand (
  RIntact s5)                (Gand (
  RIntact s6)                (
  RIntact s7)))).

(* intact caller saved registers except ...**)
Definition ET0T1T2 : action := 
  Gand (RIntact t3) (Gand (RIntact t4) (Gand (RIntact t5) (
  Gand (RIntact t6) (Gand (RIntact t7) (Gand (RIntact t8) (
       (RIntact t9))))))).

Definition POPS0S1S2 : action := 
  fun (s s' : state) =>
    getValinR s' s0 = getVinH s (getValinR s fp - 1) /\
    getValinR s' s1 = getVinH s (getValinR s fp - 2) /\
    getValinR s' s2 = getVinH s (getValinR s fp - 3) .

Definition RESRqd : action :=
  fun (s s' : state) =>
    getValinR s' fp = getVinH s (getValinR s fp) /\
    getValinR s' sp = getValinR s sp + 4 /\
    getValinR s' ra = getValinR s ra.

(* guarantee required *)     
Definition Rqd : action :=
  Gand (RIntact sp) (Gand (RIntact fp) (RIntact ra)).
(*** constant notation ***)
Notation header_size := 2.

(*** code for library function free **)
Notation L_free_prolog := 1.
Notation L_free_iter := 2.
Notation L_free_next := 3.
Notation L_free_trylo := 4.
Notation L_free_tryhi := 5.
Notation L_free_notjhi := 6.
Notation L_free_entry := 7.
Notation L_free_lnklo := 8.
Notation L_free_epilog := 9.
Notation L_free_notjlo := 10.

(** the entry of free **)
Notation free_ptr := a0.
Notation hp := s0.
Notation prev := s1.
Notation p := s2.
Parameter flist : addr.
Parameter TOP : addr.
Notation free_framesize := 4. (* 10 dwords *)
Notation nxt := 0. (*next field of info chunk *) 
Notation sz := 1. (*size field of info chunk *)

Definition C_free_prolog : iseq := seq (
  subiu sp sp free_framesize) (seq (
  sw sp free_framesize fp)    (seq (
  addiu fp sp free_framesize) (seq (
  sw_ fp 1 s0)                (seq (
  sw_ fp 2 s1)                (seq (
  sw_ fp 3 s2)                (
  j L_free_entry)             ))))).

Definition C_free_entry : iseq := seq(
  subiu hp free_ptr header_size) (seq (
  addiu prev zero pnil)       (seq (
  lw p zero flist)               (
  j L_free_iter                ))).

Definition C_free_iter : iseq := seq (
  bgtz p L_free_next)           (seq (
  sw hp nxt p)                  (
  j L_free_trylo                )).

Definition C_free_next : iseq := seq (
  subu t0 p hp)                 (seq (
  bgtz t0 L_free_tryhi)         (seq (
  addiu prev p 0)               (seq (
  lw p p nxt)                   (
  j L_free_iter                  )))).

(** merge with higher possible free block, if no , branch to L_free_notjhi **)
Definition C_free_tryhi : iseq := seq (
  lw t0 hp sz)                   (seq (
  addu t1 hp t0)                 (seq (
  bne p t1 L_free_notjhi)        (seq (
  lw t1 p sz)                    (seq (
  addu t1 t1 t0)                 (seq (
  sw hp sz t1)                   (seq (
  lw t0 p nxt)                   (seq (
  sw hp nxt t0)                  (
  j L_free_trylo                 )))))))).

Definition C_free_notjhi : iseq := seq (
  sw hp nxt p)                         (
  j L_free_trylo).

Definition C_free_trylo : iseq := seq (
  bgtz prev L_free_lnklo)  (seq (
  sw zero flist hp)            (
  j L_free_epilog)             ).

Definition C_free_lnklo : iseq := seq (
  lw t0 prev sz)                 (seq (
  addu t1 prev t0)               (seq (
  bne hp t1 L_free_notjlo)       (seq (
  lw t2 hp sz)                   (seq (
  addu t1 t0 t2)                 (seq (
  lw t0 hp nxt)                  (seq (
  sw prev sz t1)                 (seq (
  sw prev nxt t0)                (
  j L_free_epilog)               ))))))).

Definition C_free_notjlo : iseq := seq (
  sw prev nxt hp)                 (
  j L_free_epilog).

Definition C_free_epilog : iseq := seq (
  lw_ s2 fp 3)                    (seq (
  lw_ s1 fp 2)                    (seq (
  lw_ s0 fp 1)                    (seq (
  lw fp fp 0)                    (seq (
  addiu sp sp free_framesize)     (
  jr ra)                          )))).

(** Types of iseq in function free ***)
Definition P_free_prolog : assert :=
  fun s : state => 
    (getValinR s sp < getValinR s fp /\ getValinR s sp > 5 /\ getValinR s fp < TOP)
    /\ getValinR s ra <> 0 /\
    exists flistv : addr,
      exists lnk : addr,
	exists siz : nat, 
	  (star (MBlk (getValinR s free_ptr) lnk siz) (
	    star (pto flist flistv) (
	      star (Good flistv) (
		star (Stk TOP) arbh)))) s.


Definition G_free_prolog : action :=
  fun s : state =>
    fun s' : state =>
      (Gand S0S1S2 (Gand ET0T1T2 Rqd)) s s' /\ HIntact TOP (getValinR s sp) s s'
      /\ (
      forall others : assert, 
	(forall (h : heap) (rf rf' : rfile), others (h, rf) -> others (h, rf')) ->
	(exists flistv : addr, exists lnk : addr, exists siz : nat,
	  star (MBlk (getValinR s free_ptr) lnk siz) (
	    star (pto flist flistv) (
	      star (Good flistv) (
		star (Stk TOP) others))) s)
      ->
      exists flistv : addr,
	(star (pto flist flistv) (
	  star (Good flistv) (
	    star (Stk TOP) others))) s').
  
Definition T_free_prolog : codespec := cdspec P_free_prolog G_free_prolog.

Definition P_free_entry : assert := 
  fun s : state =>
    (getValinR s sp + free_framesize = getValinR s fp /\ getValinR s fp > 4
      /\ getValinR s fp < TOP) /\ getValinR s ra <> 0
    /\
    exists flistv : addr,
      exists lnk : addr,
	exists siz : nat,
	  (star (MBlk (getValinR s free_ptr) lnk siz) (
	    star (pto flist flistv) (
	      star (Good flistv) (
		star (Stk TOP) arbh)))) s.

Definition G_free_entry : action :=
  fun s : state =>
    fun s' : state => Gand ES0S1S2 ET0T1T2 s s' /\
      getValinR s' s0 = getVinH s (getValinR s fp - 1) /\
      getValinR s' s1 = getVinH s (getValinR s fp - 2) /\
      getValinR s' s2 = getVinH s (getValinR s fp - 3) /\
      getValinR s' fp = getVinH s (getValinR s fp) /\
      getValinR s' sp = getValinR s sp + 4 /\
      getValinR s' ra = getValinR s ra /\
      HIntact TOP (getValinR s fp) s s'/\ 
      (forall others : assert, 
	(forall (h : heap) (rf rf' : rfile), others (h, rf) -> others (h, rf')) 
	-> 
	(exists flistv : addr, exists lnk : addr, exists siz : nat,
	  star (MBlk (getValinR s free_ptr) lnk siz) (
	    star (pto flist flistv) (
	      star (Good flistv) (
		star (Stk TOP) others))) s)
	->
	exists flistv : addr,
	  (star (pto flist flistv) (
	    star (Good flistv) (
	      star (Stk TOP) others))) s').

Definition T_free_entry : codespec := 
  cdspec P_free_entry G_free_entry.

Definition P_free_iter : assert :=
  fun s : state => (getValinR s sp + free_framesize = getValinR s fp /\ getValinR s fp > 4
    /\ getValinR s fp < TOP /\ getValinR s prev < getValinR s hp) /\ getValinR s ra <> 0 /\
  exists flistv : addr, exists lnk : addr, exists siz : nat,
    (star (MBlk ((getValinR s hp) + 2) lnk siz) (
      star (pto flist flistv) (
	star (MidL flistv (getValinR s prev) (getValinR s p)) (
	  star (Stk TOP) arbh)))) s.
Definition G_free_iter : action :=
  fun s : state => 
    fun s' : state => 
      POPS0S1S2 s s' /\ Gand ES0S1S2 ET0T1T2 s s' /\
      getValinR s' fp = getVinH s (getValinR s fp) /\
      getValinR s' sp = getValinR s sp + 4 /\
      getValinR s' ra = getValinR s ra /\
      HIntact TOP (getValinR s fp) s s' /\
      forall others : assert,
	(forall (h : heap) (rf rf' : rfile), others (h, rf) -> others (h, rf')) ->
	(exists flistv : addr, exists lnk : addr, exists siz : nat, 	
	  (star (MBlk ((getValinR s hp) + 2) lnk siz) (
	    star (pto flist flistv) (
	      star (MidL flistv (getValinR s prev) (getValinR s p)) (
		star (Stk TOP) others)))) s) ->
	(exists flistv : addr, (star (pto flist flistv) (
	  star (Good flistv) (
	  star (Stk TOP) others))) s').

Definition T_free_iter := cdspec P_free_iter G_free_iter.

Definition P_free_next : assert :=
  fun s : state => getValinR s sp + free_framesize = getValinR s fp /\ getValinR s fp > 4
    /\ getValinR s fp < TOP /\ getValinR s prev < getValinR s hp /\ getValinR s p <> pnil 
    /\ getValinR s ra <> 0/\
  exists flistv : addr, exists lnk : addr, exists siz : nat,
    (star (MBlk ((getValinR s hp) + 2) lnk siz) (
      star (pto flist flistv) (
	star (MidL flistv (getValinR s prev) (getValinR s p)) (
	  star (Stk TOP) arbh)))) s.
Definition G_free_next : action :=
  fun s : state => 
    fun s' : state => 
      POPS0S1S2 s s' /\ Gand ES0S1S2 ET0T1T2 s s' /\
      getValinR s' fp = getVinH s (getValinR s fp) /\
      getValinR s' sp = getValinR s sp + 4 /\
      getValinR s' ra = getValinR s ra /\
      HIntact TOP (getValinR s fp) s s' /\
      forall others : assert,
	(forall (h : heap) (rf rf' : rfile), others (h, rf) -> others (h, rf')) ->
	(exists flistv : addr, exists lnk : addr, exists siz : nat, 	
	  (star (MBlk ((getValinR s hp) + 2) lnk siz) (
	    star (pto flist flistv) (
	      star (MidL flistv (getValinR s prev) (getValinR s p)) (
		star (Stk TOP) others)))) s) ->
	(exists flistv : addr, (star (pto flist flistv) (
	  star (Good flistv) (
	  star (Stk TOP) others))) s').

Definition T_free_next := cdspec P_free_next G_free_next.

Definition P_free_trylo : assert :=
  fun s : state => getValinR s sp + free_framesize = getValinR s fp /\ getValinR s fp > 4
    /\ getValinR s fp < TOP /\ getValinR s prev < getValinR s hp /\ getValinR s ra <> 0 /\
    exists flistv : addr, exists lnk : addr, exists n : nat,
      (star (MBlkLst (S n) (getValinR s hp) pnil) (
	star (pto flist flistv) (
	  star (EndL flistv (getValinR s prev) lnk) (
	    star (Stk TOP) arbh)))) s.

Definition G_free_trylo : action :=
  fun s : state => 
    fun s' : state => 
      POPS0S1S2 s s' /\ Gand ES0S1S2 ET0T1T2 s s' /\
      getValinR s' fp = getVinH s (getValinR s fp) /\
      getValinR s' sp = getValinR s sp + 4 /\
      getValinR s' ra = getValinR s ra /\
      HIntact TOP (getValinR s fp) s s' /\
      forall others : assert,
	(forall (h : heap) (rf rf' : rfile), others (h, rf) -> others (h, rf')) ->
	(exists flistv : addr, exists lnk : addr, exists n : nat, 	
	  (star (MBlkLst (S n) (getValinR s hp) pnil) (
	    star (pto flist flistv) (
	      star (EndL flistv (getValinR s prev) lnk) (
		star (Stk TOP) others)))) s) ->
	(exists flistv : addr, (star (pto flist flistv) (
	  star (Good flistv) (
	  star (Stk TOP) others))) s').

Definition T_free_trylo := cdspec P_free_trylo G_free_trylo.

Definition FREE_FRAME : assert := 
  fun s : state =>
    getValinR s sp + free_framesize = getValinR s fp /\ getValinR s fp > 4
    /\ getValinR s fp < TOP.

Definition P_free_tryhi : assert :=
  fun s : state =>
    FREE_FRAME s /\ getValinR s prev < getValinR s hp /\ getValinR s p <> pnil /\
    getValinR s hp < getValinR s p /\ getValinR s ra <> 0 /\
    exists flistv : addr, exists lnk : addr, exists siz : nat,
      (star (MBlk ((getValinR s hp) + 2) lnk siz) (
	  star (pto flist flistv) (
	    star (MidL flistv (getValinR s prev) (getValinR s p)) (
	      star (Stk TOP) arbh)))) s.

Definition G_free_tryhi : action :=
  fun s : state => 
    fun s' : state => 
      POPS0S1S2 s s' /\ Gand ES0S1S2 ET0T1T2 s s' /\ RESRqd s s' /\
      HIntact TOP (getValinR s fp) s s' /\
      forall others : assert,
	RfInsA others  ->
	(exists flistv : addr, exists lnk : addr, exists siz : nat, 	
	  (star (MBlk ((getValinR s hp) + 2) lnk siz) (
	    star (pto flist flistv) (
	      star (MidL flistv (getValinR s prev) (getValinR s p)) (
		star (Stk TOP) others)))) s) ->
	(exists flistv : addr, (star (pto flist flistv) (
	  star (Good flistv) (
	    star (Stk TOP) others))) s').

Definition T_free_tryhi := cdspec P_free_tryhi G_free_tryhi.

Definition P_free_notjhi : assert :=
  fun s : state =>
    FREE_FRAME s /\ getValinR s prev < getValinR s hp /\ getValinR s ra <> 0 /\ 
    exists flistv : addr, exists lnk : addr, exists siz : nat,
      (star (MBlk ((getValinR s hp) + 2) lnk siz) (
	star (pto flist flistv) (
	  star (MidL flistv (getValinR s prev) (getValinR s p)) (
	      star (Stk TOP) arbh)))) s /\ getValinR s p > getValinR s hp + siz.
Definition G_free_notjhi : action :=
  fun s : state => 
    fun s' : state => 
      POPS0S1S2 s s' /\ Gand ES0S1S2 ET0T1T2 s s' /\ RESRqd s s' /\
      HIntact TOP (getValinR s fp) s s' /\
      forall others : assert,
	RfInsA others  ->
	(exists flistv : addr, exists lnk : addr, exists siz : nat, 	
	  (star (MBlk ((getValinR s hp) + 2) lnk siz) (
	    star (pto flist flistv) (
	      star (MidL flistv (getValinR s prev) (getValinR s p)) (
		star (Stk TOP) others)))) s) ->
	(exists flistv : addr, (star (pto flist flistv) (
	  star (Good flistv) (
	    star (Stk TOP) others))) s').

Definition T_free_notjhi := cdspec P_free_notjhi G_free_notjhi.

Definition P_free_lnklo : assert :=
  fun s : state => FREE_FRAME s /\ getValinR s prev < getValinR s hp
    /\ getValinR s prev <> pnil /\ getValinR s ra <> 0 /\ 
    exists flistv : addr, exists lnk : addr, exists n : nat, exists m : nat, exists siz,
      (star (MBlkLst (S n) (getValinR s hp) pnil) (
	star (pto flist flistv) (
	  star (star (MBlkLst m flistv (getValinR s prev)) 
	    (MBlk (getValinR s prev + 2) lnk siz)) (
	      star (Stk TOP) arbh)))) s.

Definition G_free_lnklo : action :=
  fun s s' : state =>
    POPS0S1S2 s s' /\ Gand ES0S1S2 ET0T1T2 s s' /\ RESRqd s s' /\
    HIntact TOP (getValinR s fp) s s' /\
    forall others : assert, RfInsA others ->
      (exists flistv : addr, exists lnk : addr, exists n : nat, exists m : nat, exists siz,  	
	(star (MBlkLst (S n) (getValinR s hp) pnil) (
	  star (pto flist flistv) (
	    star (star (MBlkLst m flistv (getValinR s prev)) 
	      (MBlk (getValinR s prev + 2) lnk siz)) (
		star (Stk TOP) others)))) s) ->
      (exists flistv : addr, (star (pto flist flistv) (
	star (Good flistv) (
	  star (Stk TOP) others))) s').

Definition T_free_lnklo := cdspec P_free_lnklo G_free_lnklo.

Definition P_free_epilog : assert :=
  fun s : state => FREE_FRAME s /\ getValinR s ra <> 0 /\ 
    exists flistv : addr, 
      (star (pto flist flistv) 
	(star (Good flistv)
	  (star (Stk TOP) arbh))) s.

Definition G_free_epilog : action :=
  fun s s' : state =>
    POPS0S1S2 s s' /\ Gand ES0S1S2 ET0T1T2 s s' /\ RESRqd s s' /\
    HIntact TOP (getValinR s fp) s s' /\
    forall others : assert, RfInsA others ->
      (exists flistv : addr, (star (pto flist flistv) (star (Good flistv) 
	(star (Stk TOP) others))) s) ->
      (exists flistv : addr, (star (pto flist flistv) (star (Good flistv)
	(star (Stk TOP) others))) s').

Definition T_free_epilog := cdspec P_free_epilog G_free_epilog.

Definition P_free_notjlo : assert :=
  fun s : state => FREE_FRAME s /\ getValinR s prev < getValinR s hp
    /\ getValinR s prev <> pnil /\ getValinR s ra <> 0 /\ 
    getValinR s prev + getValinR s t0 < getValinR s hp /\
    exists flistv : addr, exists lnk : addr, exists n : nat, exists m : nat, 
      (star (MBlkLst (S n) (getValinR s hp) pnil) (
	star (pto flist flistv) (
	  star (star (MBlkLst m flistv (getValinR s prev)) 
	    (MBlk (getValinR s prev + 2) lnk (getValinR s t0))) (
	      star (Stk TOP) arbh)))) s.

Definition G_free_notjlo : action :=
  fun s s' : state =>
    POPS0S1S2 s s' /\ Gand ES0S1S2 ET0T1T2 s s' /\ RESRqd s s' /\
    HIntact TOP (getValinR s fp) s s' /\
    forall others : assert, RfInsA others ->
      (exists flistv : addr, exists lnk : addr, exists n : nat, exists m : nat,
	(star (MBlkLst (S n) (getValinR s hp) pnil) (
	  star (pto flist flistv) (
	    star (star (MBlkLst m flistv (getValinR s prev)) 
	      (MBlk (getValinR s prev + 2) lnk (getValinR s t0))) (
		star (Stk TOP) others)))) s) ->
      (exists flistv : addr, (star (pto flist flistv) (
	star (Good flistv) (
	  star (Stk TOP) others))) s').


Definition T_free_notjlo := cdspec P_free_notjlo G_free_notjlo.

