(** heaps are encoded as partial functions,
where the ordering does not matter. *)

(** This module is implemented by Dachuan Yu, September 2002 *)


Require Import EqNat.
Require  Export natutil.

(** ** syntactic categories *)

(** We use nat to simulate integers *)
Notation int := nat (only parsing).

(** heap labels *)
Notation addr := nat (only parsing).

(** heap values *)
Notation hval := nat (only parsing).

(** range of heap *)
Notation hvopt := (option hval) (only parsing).

(** short hands for options *)

(** allocated *)
Notation hvsome := Some (only parsing).

(** unallocated *)
Notation hvnone := None (only parsing).

(** nil pointer *)
Notation pnil := 0 (only parsing).

(** ** heap definitions and utility functions *)

(** heap *)
Definition heap : Set := addr -> hvopt.

(** the empty heap *)
Definition hempty : heap :=
  fun a : addr => hvnone.

(** a singleton heap *)
Definition hsingleton (l : addr) (v : hval) : heap :=
  fun a : addr => if beq_nat l a then hvsome v else hvnone.

(** equality of heaps *)
Inductive heap_eq : heap -> heap -> Prop :=
  all_same : 
  forall h1 h2 : heap, (forall i : addr, h1 i = h2 i) -> heap_eq h1 h2.

(** write to heap *)
Definition hwrite (h : heap) (l : addr) (v : hval) : heap :=
  fun a : addr => if beq_nat l a then hvsome v else h a.

(** read from heap *)
Definition hread (h : heap) (a : addr) : hvopt := h a .
Definition hread' (h : heap) (i : addr) : nat :=
  match hread h i with
    | Some n => n
    | None => 0
  end.

(** write n elements to heap, starting from address i; these n elements
  is given by function l*)
Fixpoint hwriteN (n : nat) (h : heap) (i : addr) (l : addr -> hval) {struct n} : heap :=
  match n with 
    | O => h
      | S m => hwriteN m (hwrite h i (l 0)) (S i) (fun x : addr => l (S x))
  end.

(** free a heap cell *)
Definition hfree (h : heap) (l : addr) : heap :=
  fun a : addr => if beq_nat l a then hvnone else h a.

(** free n heap cell *)
Fixpoint hfreeN (n : nat) (h : heap) (l : addr) {struct n} : heap :=
  match n with 
    | O => h
      | S m => hfreeN m (hfree h l) (S l)
  end.

(** address is not in the domain of the heap *)
Inductive hnotindom : addr -> heap -> Prop := 
  notindom_none : forall (a : addr) (h : heap), h a = hvnone -> hnotindom a h.

(** address is in the domain of the heap *)
Inductive hindom : addr -> heap -> Prop :=
  indom_some : forall (a : addr) (h : heap) (v : hval), h a = hvsome v -> hindom a h.

(** if hindom then hvsome value *)
Lemma hindom_hvsome :
  forall (a : addr) (h : heap) ,
    hindom a h -> (exists v : hval, h a = hvsome v).
Proof.
intros.
induction H.
exists v;auto.
Qed.

(** two heaps with disjoint domain *)
Inductive hdisj : heap -> heap -> Prop := 
  disj_notboth : 
  forall h1 h2 : heap, 
    (forall a : addr, hnotindom a h1 \/ hnotindom a h2) -> hdisj h1 h2.

(** merge two heaps, precondition is that (hdisj h1 h2) *)
Definition happend (h1 h2 : heap) : heap := 
  fun a : addr => match h1 a with 
		    | hvsome v => hvsome v
		      | hvnone => h2 a
		  end.

(** Properties of the heap *)

(** decidability of option types *)
Lemma opt_dec : 
  forall (A : Set) (o : option A), o = None \/ (exists x : A, o = Some x).
Proof.
simple induction o.
intros.
right.
exists a;auto.
left;auto.
Qed.

(** reflexivity of heap equality *)
Lemma heap_eq_refl : forall h : heap, heap_eq h h.
Proof.
intro.
constructor 1. (*identical to apply all_same*)
intros.
auto.
Qed.

Lemma heap_eq_sym : forall h1 h2 : heap, heap_eq h1 h2 -> heap_eq h2 h1.
Proof.
intros.
constructor 1.
intro.
induction H.
symmetry.
apply H.
Qed.

Lemma heap_eq_veq : forall (h1 h2 : heap) (i : addr), 
  heap_eq h1 h2 -> hread h1 i = hread h2 i.
Proof.
intros.
inversion_clear H.
unfold hread.
auto.
Qed.

(** transitivity of heap equality *)
Lemma heap_eq_trans : 
  forall h1 h2 h3 : heap, heap_eq h1 h2 -> heap_eq h2 h3 -> heap_eq h1 h3.
Proof.
intros.
inversion_clear H.
inversion_clear H0.
constructor 1.
intro.
replace (h3 i) with (h2 i).
apply H1.
apply H.
Qed.

Lemma heap_eq_free:
  forall (h h' : heap)(a : addr),
     heap_eq h h' -> heap_eq (hfree h a)(hfree h' a).
Proof.
intros.
inversion_clear H.
apply all_same.
intros; unfold hfree.
compare a i.
intros.
assert (true = EqNat.beq_nat a i).
apply beq_eq_true; auto.
rewrite <- H; auto.
intros.
assert (false = EqNat.beq_nat a i).
apply beq_neq_false; auto.
rewrite <- H; auto.
Qed.

Lemma heap_eq_preserve_indom :
  forall (h h' :heap)(x : addr),
    heap_eq h h' -> hindom x h -> hindom x h'.
Proof.
intuition.
inversion_clear H0.
apply indom_some with v.
inversion_clear H.
assert (h x = h' x).
apply H0.
rewrite <- H;auto.
Save.


(** empty heap is disjointed with any heaps *)
Lemma disj_emp : forall h : heap, hdisj hempty h.
Proof.
intro.
constructor 1.
intro.
left.
constructor 1.
unfold hempty;trivial.
Qed.

(** symmetry of disjointedness *)
Lemma disj_sym : forall h1 h2 : heap, hdisj h1 h2 -> hdisj h2 h1.
Proof.
intros.
inversion_clear H.
constructor.
intros.
cut (hnotindom a h1 \/ hnotindom a h2).
intro OR.
induction OR as [H3 | H3].
right;auto.
left;auto.
auto.
Qed.

(** a singleton heap containing "slot i" disjoint with a heap with "slot i" freed *)
Lemma disj_singleton_free : 
  forall (a : heap) (i : addr) (n : hval), hdisj (hsingleton i n) (hfree a i).
Proof.
intros.
constructor.
intro.
cut (false = beq_nat i a0 \/ true = beq_nat i a0).
intro OR.
induction OR.
left.
constructor.
unfold hsingleton.
rewrite <- H;auto.
right.
constructor.
unfold hfree.
rewrite <- H;auto.
generalize (beq_nat i a0).
induction b;auto.
Qed.

(*if i1 in h1, i2 in h2, and h1 disjoint with h2, then surely i1<>i2 *)
Lemma disj_indom_neq : 
  forall (h1 h2 : heap) (i1 i2 : addr), 
    hdisj h1 h2 -> hindom i1 h1 -> hindom i2 h2 -> i1 <> i2.
Proof.
intros.
inversion_clear H.
inversion_clear H0.
inversion_clear H1.
cut (hnotindom i1 h1 \/ hnotindom i1 h2);intros.
intro EQ.
rewrite <- EQ in H0.
inversion_clear H1.
inversion_clear H3.
rewrite H1 in H.
inversion H.
inversion_clear H3.
rewrite H1 in H0.
inversion H0.
apply H2.
Qed.

(** i in h implies not i not in h *)
Lemma indom_not_notin : 
  forall (h : heap) (i : addr), hindom i h -> ~ hnotindom i h.
Proof.
intros.
intro.
inversion_clear H.
inversion_clear H0.
rewrite H in H1.
inversion H1.
Qed.

(** notindom implies not indom *)
Lemma notin_not_indom : 
  forall (h : heap) (i : addr), hnotindom i h -> ~ hindom i h.
Proof.
intros.
intro.
inversion_clear H.
inversion_clear H0.
rewrite H in H1.
inversion H1.
Qed.

(** decidability of indom or not in dom *)
Lemma indom_or_notin : 
  forall (h : heap) (i : addr), hindom i h \/ hnotindom i h.
Proof.
intros.
cut (h i = None \/ exists n : nat, h i = Some n).
intro HH.
inversion_clear HH.
right; constructor;auto.
left.
inversion_clear H.
constructor 1 with x;auto.
generalize (h i).
apply opt_dec.
Qed.

(** notindom after being freed *)
Lemma free_notin : forall (h : heap) (i : addr), hnotindom i (hfree h i).
Proof.
intros.
constructor.
unfold hfree.
replace (beq_nat i i) with true;auto.
apply beq_nat_refl.
Qed.

(** notindom perserved after a free operation *)
Lemma free_preserve_notin :
  forall (h : heap) (i j : addr), hnotindom i h -> hnotindom i (hfree h j).
Proof.
intros.
inversion_clear H.
constructor.
unfold hfree.
generalize (beq_nat j i).
induction b;auto.
Qed.

(** two heap disjoint preserved after free operation *)
Lemma free_preserve_disj :
  forall (h1 h2 : heap) (i : addr), hdisj h1 h2 -> hdisj (hfree h1 i) h2.
intros.
inversion_clear H.
constructor.
intro.
cut (hnotindom a h1 \/ hnotindom a h2).
intro OR;inversion_clear OR.
left.
apply free_preserve_notin;auto.
right;auto.
auto.
Qed.

(** free a invalid heap slot doesn't affect the heap *)
Lemma notin_free_eq : 
  forall (h : heap) (i : addr), hnotindom i h -> heap_eq h (hfree h i).
Proof.
intros.
constructor.
intro.
inversion_clear H.
unfold hfree.
cut (true = beq_nat i i0 \/ false = beq_nat i i0);intros.
inversion_clear H.
rewrite <- H1.
replace i0 with i;auto.
apply beq_nat_eq;auto.
rewrite <- H1;auto.
generalize (beq_nat i i0).
induction b.
auto.
auto.
Qed.

(**i in h , after write i still in h **)
Lemma write_preserver_indom :
  forall (h : heap) (i j : addr) (v : nat),
    hindom i h -> hindom i (hwrite h j v).
Proof.
intros.
inversion_clear H.
unfold hwrite.
compare j i.
intros.
apply indom_some with v.
assert (true = beq_nat j i).
apply beq_eq_true; auto.
rewrite <- H; auto.
intros.
apply indom_some with v0.
assert (false = beq_nat j i).
apply beq_neq_false; auto.
rewrite <- H; auto.
Qed.

(** two heaps remain disjoint after writing into a trivial slot *)
Lemma notin_write_preserve_disj_R : 
  forall (h1 h2 : heap) (i : addr) (n : hval),
    hdisj h1 h2 -> hnotindom i h1 -> hdisj h1 (hwrite h2 i n).
Proof.
intros.
inversion_clear H.
inversion_clear H0.
constructor.
intros.
cut (hnotindom a h1 \/ hnotindom a h2);intros.
inversion_clear H0.
left;auto.
cut (true = beq_nat i a \/ false = beq_nat i a).
intro EQorNEQ.
inversion_clear EQorNEQ.
left.
constructor.
replace  a with i;auto.
apply beq_nat_eq;auto.
right.
constructor.
unfold hwrite.
rewrite <- H0.
inversion_clear H2;auto.
generalize (beq_nat i a).
induction b.
auto.
auto.
apply H1;auto.
Qed.

Lemma notin_write_preserve_disj_L : 
  forall (h1 h2 : heap) (i : addr) (n : hval), 
    hdisj h1 h2 -> hnotindom i h2 -> hdisj (hwrite h1 i n) h2.
Proof.
intros.
apply disj_sym.
apply notin_write_preserve_disj_R.
apply disj_sym;auto.
auto.
Qed.

(*two disjoint heaps, i in h1, then i not in h2 *)
Lemma disj_inl_notinr : 
  forall (h1 h2 : heap) (i : addr),
    hdisj h1 h2 -> hindom i h1 -> hnotindom i h2.
Proof.
intros.
inversion_clear H.
inversion_clear H0.
constructor.
cut (hnotindom i h1 \/ hnotindom i h2);intros.
inversion_clear H0.
inversion_clear H2.
rewrite H0 in H.
inversion_clear H.
inversion_clear H2;auto.
apply H1;auto.
Qed.

Lemma disj_inr_notinl :
  forall (h1 h2 : heap) (a : addr), 
    hdisj h1 h2 -> hindom a h2 -> hnotindom a h1.
Proof.
intros.
apply disj_inl_notinr with h2.
apply disj_sym;auto.
auto.
Qed.

Lemma in_write_preserve_disj_L : 
  forall (h1 h2 : heap) (a : addr) (n : hval),
    hdisj h1 h2 -> hindom a h1 -> hdisj (hwrite h1 a n) h2.
Proof.
intros.
apply notin_write_preserve_disj_L.
auto.
apply disj_inl_notinr with h1;auto.
Qed.

(** appending a empty heap doesn't affect the heap *)
Lemma append_emp : forall h : heap, heap_eq (happend hempty h) h.
Proof.
intros.
unfold hempty.
unfold happend.
constructor.
auto.
Qed.

(** symmetry of appending *)
Lemma append_sym : 
  forall h1 h2 : heap, hdisj h1 h2 -> heap_eq (happend h1 h2) (happend h2 h1).
Proof.
intros.
constructor.
intros.
inversion_clear H.
cut (hnotindom i h1 \/ hnotindom i h2);intros.
inversion_clear H.
inversion_clear H1.
unfold happend.
rewrite H.
generalize (h2 i).
induction o;auto.
inversion_clear H1.
unfold happend.
rewrite H.
generalize (h1 i).
induction o;auto.
apply H0;auto.
Qed.

Lemma append_assoc : forall(h h' h'':heap),
  hdisj h h'->hdisj h h''->hdisj h' h''->
  heap_eq (happend h (happend h' h'')) (happend (happend h h') h'').
Proof.
unfold happend;intuition.
inversion_clear H;inversion_clear H0;inversion_clear H1.
constructor;intros.
cut(hnotindom i h\/hnotindom i h');[intros|apply H2].
cut(hnotindom i h\/hnotindom i h'');[intros|apply H].
cut(hnotindom i h'\/hnotindom i h'');[intros|apply H0].
intuition;inversion_clear H5;inversion_clear H1;inversion_clear H3;
induction (h i);induction (h' i);induction (h'' i);try discriminate;auto.
Save.



(** two identical heap , i in h1 then i ni h2 too *)
Lemma h_subst_indom :
  forall (h1 h2 : heap) (i : addr), heap_eq h1 h2 -> hindom i h1 -> hindom i h2.
Proof.
intros.
inversion_clear H.
inversion_clear H0.
constructor 1 with v.
replace (h2 i) with (h1 i);auto.
Qed.

(** append don't change the value of another heap *)
Lemma append_preserve_read_R : 
  forall (h1 h2 : heap) (i : addr),
    hnotindom i h1 -> hread (happend h1 h2) i = hread h2 i.
Proof.
intros.
inversion_clear H.
unfold hread.
unfold happend.
rewrite H0;auto.
Qed.

Lemma append_preserve_read_L :
  forall (h1 h2 : heap) (i : addr), 
    hnotindom i h2 -> hread (happend h1 h2) i = hread h1 i.
Proof.
intros.
inversion_clear H.
unfold hread.
unfold happend.
rewrite H0.
generalize (h1 i).
induction o;auto.
Qed.

(** i in one heap, then i in the new heap after appending *)
Lemma append_preserve_indom_L :
  forall (h1 h2 : heap) (i : addr), hindom i h1 -> hindom i (happend h1 h2).
Proof.
intros.
inversion_clear H.
constructor 1 with v.
unfold happend.
rewrite H0;auto.
Qed.

Lemma append_perserve_indom_R :
  forall (h1 h2 : heap) (i : addr), 
    hindom i h2 -> hdisj h1 h2 -> hindom i (happend h1 h2).
Proof.
intros.
inversion H.
constructor 1 with v.
unfold happend.
replace (h1 i) with (None:option nat);auto.
cut (hnotindom i h1);intros.
inversion_clear H4.
rewrite H5;auto.
apply disj_inr_notinl with h2;auto.
Qed.

(** if both notin domain, then notin after appending *)
Lemma append_preserve_notin : 
  forall (h1 h2 : heap) (i : addr),
    hnotindom i h1 -> hnotindom i h2 -> hnotindom i (happend h1 h2).
Proof.
intros.
constructor.
unfold happend.
inversion_clear H.
inversion_clear H0.
rewrite H1;auto.
Qed.

(** similar to disj_inl_notinr *)
Lemma disj_found_notin :
  forall (h1 h2 : heap) (i : addr) (n : hval),
    hdisj h1 h2 -> hread h1 i = Some n -> hnotindom i h2.
intros.
cut (hindom i h1);intros.
apply disj_inl_notinr with h1;auto.
constructor 1 with n.
unfold hread in H0;auto.
Qed.




Lemma disj_append_L : forall (h h' h'':heap),
  (hdisj h h' /\ hdisj h h'') -> hdisj h (happend h' h'').
Proof.
intuition.
apply disj_notboth;intuition.
inversion_clear H0.
inversion_clear H1.
unfold happend.
cut(hnotindom a h \/ hnotindom a h');[intros|apply H].
cut(hnotindom a h \/ hnotindom a h'');[intros|apply H0].
intuition;right.
inversion_clear H3;inversion_clear H1.
apply notindom_none. 
rewrite H2;auto.
Save.

Lemma disj_append_R : forall (h h' h'':heap),
  hdisj h (happend h' h'') -> (hdisj h h' /\ hdisj h h'').
Proof.
intuition.
constructor.
inversion_clear H;intuition.
cut(hnotindom a h \/ hnotindom a (happend h' h''));[intuition;right|apply H0].
constructor.
inversion_clear H1.
unfold happend in H;intuition.
induction (h' a);[discriminate|auto].
constructor.
inversion_clear H;intuition.
cut(hnotindom a h \/ hnotindom a (happend h' h''));[intuition;right|apply H0].
constructor.
inversion_clear H1.
unfold happend in H;intuition.
induction (h'' a);induction (h' a);try discriminate;auto.
Save.

Lemma disj_heap_eq :forall(h h' h'':heap),
  hdisj h h' -> heap_eq h' h'' -> hdisj h h''.
Proof.
intros.
constructor;intuition.
inversion_clear H.
inversion_clear H0.
cut (h' a = h'' a);[intros|apply H].
cut (hnotindom a h \/ hnotindom a h');[intros|apply H1].
inversion_clear H2;intuition.
inversion_clear H3.
right;constructor.
rewrite H0 in H2;auto.
Save.

Lemma heap_eq_append :forall(h h' h'':heap),
  heap_eq h' h''->heap_eq (happend h' h) (happend h'' h).
Proof.
intros;constructor;intros.
unfold happend;inversion_clear H;intuition.
cut(h' i = h'' i);[intros|apply H0].
rewrite H;tauto.
Save.

Lemma append_nullheap :forall(h h':heap),
  heap_eq hempty h -> heap_eq (happend h h') h'.
Proof.
constructor;intros.
unfold happend;unfold hempty in H;inversion_clear H;intuition.
cut(h i = None);[intros|symmetry;apply H0].
rewrite H;auto.
Save.

Lemma indom_part_R :
  forall (h h' h'' : heap) (i : nat),
    hindom i h -> hdisj h' h'' -> heap_eq (happend h' h'') h -> 
    hnotindom i h' -> hindom i h''.
Proof.
intros.
inversion_clear H.
apply indom_some with v.
inversion_clear H1.
inversion_clear H2.
assert (happend h' h'' i = h i);auto.
unfold happend in H2;rewrite H3 in H2;rewrite H1 in H2;simpl;auto.
Save.


(** Add to "fsl" hints database *)
Hint Resolve beq_neq_false : fsl.
(*Print Hint beq_neq_false.*)
Hint Immediate disj_emp append_preserve_notin : fsl.
(** While met with a goal like "hread (happend ? ?) ? = Some ?", *)
(** if auto can't succeed using a tactic with a cost less than 1,using below*)
(** the goal has a hypothesis that ?X2 accomodates ?X4 *)
(*Check append_preserve_read_R.*)
(*Print disj_found_notin*)
Hint Extern 1 (hread (happend ?X1 ?X2) ?X3 = Some ?X4) =>
  match goal with 
    |rw : (hread ?X2 ?X3 = Some ?X4) |- (hread (happend ?X1 ?X2) ?X3 = Some ?X4)
      =>rewrite <- rw;apply append_preserve_read_R;
	apply disj_found_notin with X2 X4;
	  [ apply disj_sym; assumption | assumption ]
  end : fsl.

      
(*Print HintDb fsl.*)



