Require Import EqNat.
Require Import natutil.
Require Import heap2.
Require Import Omega.

(** ** Syntactic Categories & Related Utilities *)

(** code and function labels *)
Notation lab := nat (only parsing).

(** register *)
Inductive reg : Set :=
  | r0 : reg | r1 : reg | r2 : reg | r3 : reg | r4 : reg | r5 : reg | r6 : reg 
  | r7 : reg | r8 : reg | r9 : reg | r10 : reg | r11 : reg | r12 : reg | r13 : reg
  | r14 : reg | r15 : reg | r16 : reg | r17 : reg | r18 : reg | r19 : reg | r20 : reg 
  | r21 : reg | r22 : reg | r23 : reg | r24 : reg | r25 : reg | r26 : reg | r27 : reg
  | r28 : reg | r29 : reg | r30 : reg | r31 : reg.

(*** Registers alias of convention **)
Notation ra := r31. (* return address proc *)
Notation zero := r0. (*always zero *)
Notation ast := r1. (* assembler temporary *)
Notation v0 := r2. (*v0 - v1 return values *)
Notation v1 := r3.
Notation a0 := r4. (*a0 - a3 arguments *)
Notation a1 := r5.
Notation a2 := r6.
Notation a3 := r7.
Notation t0 := r8. (*t0 - t9 caller saved temporary *)
Notation t1 := r9.
Notation t2 := r10. 
Notation t3 := r11.
Notation t4 := r12.
Notation t5 := r13.
Notation t6 := r14.
Notation t7 := r15.
Notation t8 := r24.
Notation t9 := r25.
Notation s0 := r16. (*s0 - s7 callee saved temporary *)
Notation s1 := r17.
Notation s2 := r18.
Notation s3 := r19.
Notation s4 := r20.
Notation s5 := r21.
Notation s6 := r22.
Notation s7 := r23.
Notation k0 := r26.
Notation k1 := r27.
Notation gp := r28.
Notation sp := r29.
Notation fp := r30.


Definition beq_reg (ra rb : reg) :=
  match ra, rb with
    | r0, r0 => true
    | r1, r1 => true
    | r2, r2 => true
    | r3, r3 => true
    | r4, r4 => true
    | r5, r5 => true
    | r6, r6 => true
    | r7, r7 => true
    | r8, r8 => true
    | r9, r9 => true
    | r10, r10 => true
    | r11, r11 => true
    | r12, r12 => true
    | r13, r13 => true
    | r14, r14 => true
    | r15, r15 => true
    | r16, r16 => true
    | r17, r17 => true
    | r18, r18 => true
    | r19, r19 => true
    | r20, r20 => true
    | r21, r21 => true
    | r22, r22 => true
    | r23, r23 => true
    | r24, r24 => true
    | r25, r25 => true
    | r26, r26 => true
    | r27, r27 => true
    | r28, r28 => true
    | r29, r29 => true
    | r30, r30 => true
    | r31, r31 => true
    | _, _ => false
  end.

Inductive rfile : Set := 
  | emptyR : rfile
  | consR : rfile -> reg -> nat -> rfile.

(** lookup reg file *)
Fixpoint lookupR (rf : rfile) (r : reg) {struct rf} : nat :=
  match rf with
    | emptyR => O
    | consR rf' r' i => if beq_reg r r' then i else lookupR rf' r
  end.

Definition updateR (rf : rfile) (r : reg) (i : nat) : rfile :=
  consR rf r i.

(** instruction set *)
Inductive iota : Set :=
  | addu : reg -> reg -> reg -> iota
  | addiu : reg -> reg -> nat -> iota
  | subu : reg -> reg -> reg -> iota
  | subiu : reg -> reg -> nat -> iota
  | beq : reg -> reg -> lab -> iota
  | bgez : reg -> lab -> iota
  | bgtz : reg -> lab -> iota
  | blez : reg -> lab -> iota
  | bltz : reg -> lab -> iota
  | bne : reg -> reg -> lab -> iota
  | lw : reg -> reg -> nat -> iota
  | lw_ : reg -> reg -> nat -> iota
  | sw : reg -> nat -> reg -> iota
  | sw_ : reg -> nat -> reg -> iota.

Inductive iseq : Set :=
  | seq : iota -> iseq -> iseq
  | j : lab -> iseq
  | jal : lab -> lab -> iseq
  | jalr : reg -> reg -> lab -> iseq
  | jr : reg -> iseq.

Notation state := (heap * rfile) % type.

Definition getValinH (s : state) (i : nat) : hvopt :=
  let (h, _ ) := s in (hread h i).

Definition getValinR (s : state) (r : reg) : nat :=
  let (_, rf) := s in
      lookupR rf r.

(** free the heap slot in state *)
Definition removeS (s : state) (i : addr ) :=
  let (h, rf) := s in (hfree h i, rf).

(*** code heap **)
Inductive codeheap : Set :=
  | emptyCH : codeheap
  | consCH : lab -> iseq -> codeheap -> codeheap.

(*** code heap utilities *)

Inductive lookupCH : codeheap -> lab -> iseq -> Prop :=
  | lookupCH_eq :
    forall (ch : codeheap) (l : lab) (b : iseq),
      lookupCH (consCH l b ch) l b
  | lookupCH_neq :
    forall (ch : codeheap) (l l' : lab) (b b' : iseq),
      lookupCH ch l' b' -> l <> l' -> lookupCH (consCH l b ch) l' b'.


(*** world : state * codeheap * iseq ***)
Notation world := (state * (codeheap * iseq)) % type.

Inductive next : iota -> state -> state -> Prop :=
  | next_addu :
    forall (rd rs rt : reg) (h : heap) (rf : rfile),
      next (addu rd rs rt)
      (h, rf)
      (h, updateR rf rd ((lookupR rf rs) + (lookupR rf rt)))
  | next_addiu :
    forall (rd rs : reg) (i : nat) (rf : rfile) (h : heap),
      next (addiu rd rs i)
      (h, rf)
      (h, updateR rf rd ((lookupR rf rs) + i))
  | next_subu :
    forall (rd rs rt : reg) (h : heap) (rf : rfile),
      next (subu rd rs rt)
      (h, rf)
      (h, updateR rf rd ((lookupR rf rs) - (lookupR rf rt)))
  | next_subiu :
    forall (rd rs : reg) (i : nat) (rf : rfile) (h : heap),
      next (subiu rd rs i)
      (h, rf)
      (h, updateR rf rd ((lookupR rf rs) - i))
  | next_lw :
    forall (rd rs : reg) (x i : nat) (h : heap) (rf : rfile),
      hread h (lookupR rf rs + i) = hvsome x -> 
      next (lw rd rs i)
      (h, rf)
      (h, updateR rf rd x)
  | next_lw_ :
    forall (rd rs : reg) (x i : nat) (h : heap) (rf : rfile),
      hread h ((lookupR rf rs) - i) = hvsome x -> 
      next (lw_ rd rs i)
      (h, rf)
      (h, updateR rf rd x)
  | next_sw : 
    forall (rd rs : reg) (i : nat) (h : heap) (rf : rfile),
      hindom (lookupR rf rd + i) h ->
      next (sw rd i rs)
      (h, rf)
      (hwrite h (lookupR rf rd + i) (lookupR rf rs), rf)
  | next_sw_ : 
    forall (rd rs : reg) (i : nat) (h : heap) (rf : rfile),
      hindom ((lookupR rf rd) - i) h ->
      next (sw_ rd i rs)
      (h, rf)
      (hwrite h ((lookupR rf rd) - i) (lookupR rf rs), rf).



Lemma stateeq_hreq :
  forall (h1 h2 : heap) (rf1 rf2 : rfile),
    (h1, rf1) = (h2, rf2) -> h1 = h2 /\ rf1 = rf2.
Proof.
intros.
inversion H.
intuition.
Save.

Lemma next_state_eq :
  forall (c : iota) (s s'1 s'2 : state),
    next c s s'1 -> next c s s'2 -> s'1 = s'2.
Proof.
intros.
inversion H.
rewrite <- H1 in H0.
inversion H0.
rewrite <- H2 in H4.
inversion H4.
auto.
rewrite <- H1 in H0.
inversion H0.
rewrite <- H2 in H4.
inversion H4.
auto.
rewrite <- H1 in H0.
inversion H0.
rewrite <- H2 in H4.
inversion H4.
auto.
rewrite <- H1 in H0.
inversion H0.
rewrite <- H2 in H4.
inversion H4.
auto.
(* lw rd rs i *)
rewrite <- H2 in H0.
inversion H0.
rewrite <- H3 in H6.
inversion H6.
rewrite H12 in H10; rewrite H13 in H10.
rewrite H1 in H10.
inversion H10; auto.
rewrite <- H2 in H0.
inversion H0.
rewrite <- H3 in H6; inversion H6.
rewrite H12 in H10; rewrite H13 in H10.
rewrite H1 in H10.
inversion H10; auto.
rewrite <- H2 in H0.
inversion H0.
rewrite <- H3 in H6.
inversion H6.
auto.
rewrite <- H2 in H0.
inversion H0.
rewrite <- H3 in H6.
inversion H6.
tauto.
Save.

Inductive step : world -> world -> Prop :=
  | step_j : 
    forall (f : lab) (h : heap) (rf : rfile) (ch : codeheap) (is : iseq),
      lookupCH ch f is ->
      step ((h, rf), (ch, j f)) ((h, rf), (ch, is))
  | step_jal :
    forall (f fret : lab) (h : heap) (rf : rfile) (ch : codeheap) (is : iseq),
      lookupCH ch f is ->
      step ((h, rf), (ch, jal f fret))
      ((h, updateR rf ra fret), (ch, is))
  | step_jalr :
    forall (fret : lab) (h : heap) (rf : rfile) (rd rs : reg) 
      (ch : codeheap) (is : iseq),
      lookupCH ch (lookupR rf rs) is ->
      step ((h, rf), (ch , jalr rd rs fret))
      ((h, updateR rf rd fret), (ch, is))
  | step_jr :
    forall (rs : reg) (ch : codeheap) (is : iseq) (h : heap) (rf : rfile),
      lookupCH ch (lookupR rf rs) is ->
      step ((h, rf), (ch, jr rs))
      ((h, rf), (ch, is))
  | step_beq_f :
    forall (rs rt : reg) (f : lab) (is : iseq) (h : heap) (rf : rfile) (ch : codeheap),
      lookupR rf rs <> lookupR rf rt ->
      step ((h, rf), (ch, seq (beq rs rt f) is))
      ((h,rf), (ch, is))
  | step_beq_t :
    forall (rs rt : reg) (f : lab) (is is' : iseq) 
      (h : heap) (rf : rfile) (ch : codeheap),
      lookupR rf rs = lookupR rf rt ->
      lookupCH ch f is' ->
      step ((h, rf), (ch, seq (beq rs rt f) is))
      ((h, rf), (ch, is'))
  | step_bne_f :
    forall (rs rt : reg) (f : lab) (is : iseq) (h : heap) (rf : rfile) (ch : codeheap),
      lookupR rf rs = lookupR rf rt ->
      step ((h, rf), (ch, seq (bne rs rt f) is))
      ((h,rf), (ch, is))
  | step_bne_t :
    forall (rs rt : reg) (f : lab) (is is' : iseq) 
      (h : heap) (rf : rfile) (ch : codeheap),
      lookupR rf rs <> lookupR rf rt ->
      lookupCH ch f is' ->
      step ((h, rf), (ch, seq (bne rs rt f) is))
      ((h, rf), (ch, is'))
  | step_bgez_f :
    forall (rs : reg) (f : lab) (is : iseq)
      (h : heap) (rf : rfile) (ch : codeheap),
      lookupR rf rs < 0 ->
      step ((h, rf), (ch, seq (bgez rs f) is))
      ((h, rf), (ch, is))
  | step_bgez_t :
    forall (rs : reg) (f : lab) (is is' : iseq)
      (h : heap) (rf : rfile) (ch : codeheap),
      lookupR rf rs >= 0 ->
      lookupCH ch f is' ->
      step ((h, rf), (ch, seq (bgez rs f) is))
      ((h, rf), (ch, is'))
  | step_bgtz_f :
    forall (rs : reg) (f : lab) (is : iseq)
      (h : heap) (rf : rfile) (ch : codeheap),
      lookupR rf rs <= 0 ->
      step ((h, rf), (ch, seq (bgtz rs f) is))
      ((h, rf), (ch, is))
  | step_bgtz_t :
    forall (rs : reg) (f : lab) (is is' : iseq)
      (h : heap) (rf : rfile) (ch : codeheap),
      lookupR rf rs > 0 ->
      lookupCH ch f is' ->
      step ((h, rf), (ch, seq (bgtz rs f) is))
      ((h, rf), (ch, is'))
  | step_blez_f :
    forall (rs : reg) (f : lab) (is : iseq)
      (h : heap) (rf : rfile) (ch : codeheap),
      lookupR rf rs > 0 ->
      step ((h, rf), (ch, seq (blez rs f) is))
      ((h, rf), (ch, is))
  | step_blez_t :
    forall (rs : reg) (f : lab) (is is' : iseq)
      (h : heap) (rf : rfile) (ch : codeheap),
      lookupR rf rs <= 0 ->
      lookupCH ch f is' ->
      step ((h, rf), (ch, seq (blez rs f) is))
      ((h, rf), (ch, is'))
  | step_bltz_f :
    forall (rs : reg) (f : lab) (is : iseq)
      (h : heap) (rf : rfile) (ch : codeheap),
      lookupR rf rs >= 0 ->
      step ((h, rf), (ch, seq (bltz rs f) is))
      ((h, rf), (ch, is))
  | step_bltz_t :
    forall (rs : reg) (f : lab) (is is' : iseq)
      (h : heap) (rf : rfile) (ch : codeheap),
      lookupR rf rs < 0 ->
      lookupCH ch f is' ->
      step ((h, rf), (ch, seq (bltz rs f) is))
      ((h, rf), (ch, is'))
  | step_rest :
    forall (s s' : state) (ch : codeheap) (i : iota) (is : iseq),
      next i s s' ->
      step (s, (ch, seq i is)) (s', (ch, is)).
   

      



      
