(** Static Semantics of SCAP *)

Require Import EqNat.
Require Import natutil.
Require Import heap2.
Require Import scap.

(** **Verification Structure ***)
Notation assert := (state -> Prop).

Notation action := (state -> state -> Prop).

Inductive codespec : Type := 
  cdspec : assert -> action -> codespec.

Inductive codeheapty : Type :=
  | emptyCT : codeheapty
  | consCT : lab -> codespec -> codeheapty -> codeheapty.


Inductive lookupCT : codeheapty -> lab -> codespec -> Prop :=
  | lookupCT_eq : forall (ct : codeheapty) (l : lab) (c : codespec),
    lookupCT (consCT l c ct) l c
  | lookupCT_neq : forall (ct : codeheapty) (l l' : lab) (c c' : codespec),
    lookupCT ct l' c' -> l <> l' -> lookupCT (consCT l c ct) l' c'.

(** subseteq and lemmas *)
Inductive psi_subseteq : codeheapty -> codeheapty -> Prop :=
  | left_in_right :
    forall (ct1 : codeheapty) (ct2 : codeheapty),
      (forall (l : lab) (theta : codespec),
	lookupCT ct1 l theta -> lookupCT ct2 l theta)
      -> psi_subseteq ct1 ct2.

(*subset refl*)
Lemma psi_eq_subseteq : forall (ct : codeheapty), psi_subseteq ct ct.
Proof.
intros.
constructor.
intros.
assumption.
Save.

(*subset trans*)
Lemma psi_subseteq_trans : forall (ct1 ct2 ct3 : codeheapty), 
  psi_subseteq ct1 ct2 -> psi_subseteq ct2 ct3 -> psi_subseteq ct1 ct3.
Proof.
intros.
constructor;intros.
inversion_clear H;inversion_clear H0.
apply H;apply H2;auto.
Save.

(** Interpreted precondition **)
Notation interp := (codeheapty -> state -> Prop).

Notation NULL := 0.

Inductive WFST : nat -> assert -> codeheapty -> Prop :=
  | WFST_0 : 
    forall (q : assert) (psi : codeheapty),
      (forall (h : heap) (rf : rfile), q (h, rf) -> lookupR rf ra = NULL)
      -> (WFST O q psi)
  | WFST_n :
    forall (n : nat) (q : assert) (psi : codeheapty),
      n > 0 ->
      (forall (h : heap) (rf : rfile), 
	q (h, rf) -> (exists p' : assert, exists g' : action,  
	  lookupCT psi (lookupR rf ra) (cdspec p' g') /\ p' (h, rf)
	  /\ WFST (n - 1) (g' (h, rf)) psi)
      )
      -> WFST n q psi.

Definition itpx1 (theta : codespec) : interp :=
  match theta with
    | cdspec p g =>
      (fun psi : codeheapty => 
	fun s : state =>
	  p s /\ (exists n : nat, (WFST n (g s) psi)))
  end.

Definition itpx2 (a : interp) (psi : codeheapty) : interp :=
  fun psi' : codeheapty =>
    fun s : state => (psi_subseteq psi psi') /\ a psi' s.

Inductive Infer : interp -> iseq -> Prop :=
  | I_j : forall (a : interp) (f : lab),
    (forall (psi : codeheapty) (s : state), a psi s ->
      (exists theta : codespec, lookupCT psi f theta /\ (itpx1 theta) psi s))
    -> Infer a (j f)
  | I_jal : forall (a : interp) (f : lab) (fret : lab),
    (forall (psi : codeheapty) (h : heap) (rf : rfile), a psi (h, rf) ->
      (exists theta : codespec, lookupCT psi f theta /\ 
	(itpx1 theta) psi (h, updateR rf ra fret)))
    -> Infer a (jal f fret)
  | I_jalr : forall (a : interp) (rd rs : reg) (fret : lab),
    (forall (psi : codeheapty) (h : heap) (rf : rfile), a psi (h, rf) ->
      (exists theta : codespec, lookupCT psi (lookupR rf rs) theta /\ 
	(itpx1 theta) psi (h, updateR rf rd fret)))
    -> Infer a (jalr rd rs fret)
  | I_jr : forall (a : interp) (rs : reg),
    (forall (psi : codeheapty) (h : heap) (rf : rfile), a psi (h, rf) ->
      (exists theta : codespec, lookupCT psi (lookupR rf rs) theta /\
	(itpx1 theta) psi (h, rf)))
    -> Infer a (jr rs)
  | I_beq : forall (a a' : interp) (rs rt : reg) (f : lab) (is : iseq),
    (forall (psi : codeheapty) (h : heap) (rf : rfile), a psi (h, rf) ->
      ((lookupR rf rs <> lookupR rf rt) -> a' psi (h, rf)) /\ 
      ((lookupR rf rs = lookupR rf rt) -> 
	(exists theta : codespec, lookupCT psi f theta /\
	  (itpx1 theta) psi (h, rf))))
    -> Infer a' is
    -> Infer a (seq (beq rs rt f) is)
  | I_bne : forall (a a' : interp) (rs rt : reg) (f : lab) (is : iseq),
    (forall (psi : codeheapty) (h : heap) (rf : rfile), a psi (h, rf) ->
      ((lookupR rf rs = lookupR rf rt) -> a' psi (h, rf)) /\ 
      ((lookupR rf rs <> lookupR rf rt) -> 
	(exists theta : codespec, lookupCT psi f theta /\
	  (itpx1 theta) psi (h, rf))))
    -> Infer a' is
    -> Infer a (seq (bne rs rt f) is)
  | I_bgez : forall (a a' : interp) (rs rt : reg) (f : lab) (is : iseq),
    (forall (psi : codeheapty) (h : heap) (rf : rfile), a psi (h, rf) ->
      ((lookupR rf rs < 0) -> a' psi (h, rf)) /\ 
      ((lookupR rf rs >= 0) -> 
	(exists theta : codespec, lookupCT psi f theta /\
	  (itpx1 theta) psi (h, rf))))
    -> Infer a' is
    -> Infer a (seq (bgez rs f) is)
  | I_bgtz : forall (a a' : interp) (rs rt : reg) (f : lab) (is : iseq),
    (forall (psi : codeheapty) (h : heap) (rf : rfile), a psi (h, rf) ->
      ((lookupR rf rs <= 0) -> a' psi (h, rf)) /\ 
      ((lookupR rf rs > 0) -> 
	(exists theta : codespec, lookupCT psi f theta /\
	  (itpx1 theta) psi (h, rf))))
    -> Infer a' is
    -> Infer a (seq (bgtz rs f) is)
  | I_blez : forall (a a' : interp) (rs rt : reg) (f : lab) (is : iseq),
    (forall (psi : codeheapty) (h : heap) (rf : rfile), a psi (h, rf) ->
      ((lookupR rf rs > 0) -> a' psi (h, rf)) /\ 
      ((lookupR rf rs <= 0) -> 
	(exists theta : codespec, lookupCT psi f theta /\
	  (itpx1 theta) psi (h, rf))))
    -> Infer a' is
    -> Infer a (seq (blez rs f) is)
  | I_bltz : forall (a a' : interp) (rs rt : reg) (f : lab) (is : iseq),
    (forall (psi : codeheapty) (h : heap) (rf : rfile), a psi (h, rf) ->
      ((lookupR rf rs >= 0) -> a' psi (h, rf)) /\ 
      ((lookupR rf rs < 0) -> 
	(exists theta : codespec, lookupCT psi f theta /\
	  (itpx1 theta) psi (h, rf))))
    -> Infer a' is
    -> Infer a (seq (bltz rs f) is)
  | I_rest : forall (a a' : interp) (c : iota) (is : iseq),
    (forall (psi : codeheapty) (s : state), a psi s ->
      (exists s' : state, next c s s' /\  a' psi s'))
    -> Infer a' is
    -> Infer a (seq c is).

(** Psi |- C : Psi *)
Inductive WF_CH_0 : codeheapty -> codeheapty -> codeheap -> Prop :=
  | wf_c_emp : forall (ct0 : codeheapty),
    WF_CH_0 ct0  emptyCT emptyCH
  | wf_c_cons : forall (ct0 ct : codeheapty) (ch : codeheap)
    (f : lab) (theta : codespec) (is : iseq),
    Infer (itpx2 (itpx1 theta) ct0) is
    -> WF_CH_0 ct0 ct ch 
    -> WF_CH_0 ct0 (consCT f theta ct) (consCH f is ch).

Inductive WF_CH : codeheapty -> codeheap -> Prop :=
  | wf_c :
    forall (ct : codeheapty) (ch : codeheap),
      WF_CH_0 ct ct ch -> WF_CH ct ch.

(*** Psi; |- {a} W *)
Inductive WF_WLD : codeheapty -> interp -> world -> Prop :=
  | wf_wld :
    forall (ct : codeheapty) (a : interp) (s : state) (ch : codeheap) (is : iseq),
      a ct s 
      -> Infer a is
      -> WF_CH ct ch
      -> WF_WLD ct a (s, (ch, is)).
    






      

      
  

